@@ -32,6 +32,12 @@
 
 #define IMAGE_BAYER_NONE 0
 
+/*
+ * This constant indicates that some other default value should be filled in.
+ */
+
+#define IMAGE_BAYER_DEFAULT 0x8
+
 /*
  * Do not change these values without inspecting
  * image_bayer_ale_real::r_*_offset().
@@ -166,6 +172,10 @@ public:
 		_offset[1] = j;
 	}
 
+	void set_offset(point p) {
+		_offset = p;
+	}
+
 	unsigned int width() const {
 		return _dimx;
 	}
@@ -224,6 +234,57 @@ public:
 		return result;
 	}
 
+	/*
+	 * Get the maximum difference among adjacent pixels.
+	 */
+
+	pixel get_max_diff(unsigned int i, unsigned int j) const {
+		assert(i >= 0);
+		assert(j >= 0);
+		assert(i <= _dimy - 1);
+		assert(j <= _dimx - 1);
+
+		pixel max = get_pixel(i, j), min = get_pixel(i, j);
+
+		for (int ii = -1; ii <= 1; ii++)
+		for (int jj = -1; jj <= 1; jj++) {
+			int iii = i + ii;
+			int jjj = j + jj;
+
+			if (iii < 0)
+				continue;
+			if (jjj < 0)
+				continue;
+			if ((unsigned int) iii > _dimy - 1)
+				continue;
+			if ((unsigned int) jjj > _dimx - 1)
+				continue;
+
+			pixel p = get_pixel(iii, jjj);
+
+			for (int d = 0; d < 3; d++) {
+				if (p[d] > max[d])
+					max[d] = p[d];
+				if (p[d] < min[d])
+					min[d] = p[d];
+			}
+		}
+
+		return max - min;
+	}
+
+	pixel get_max_diff(point x) const {
+		assert (x[0] >= 0);
+		assert (x[1] >= 0);
+		assert (x[0] <= _dimy - 1);
+		assert (x[1] <= _dimx - 1);
+
+		unsigned int i = (unsigned int) round(x[0]);
+		unsigned int j = (unsigned int) round(x[1]);
+
+		return get_max_diff(i, j);
+	}
+
 	/*
 	 * Get a color value at a given position using bilinear interpolation between the
 	 * four nearest pixels.  Result values:
@@ -300,6 +361,17 @@ public:
 #endif
 	}
 
+	int in_bounds(point x) const {
+		if (x[0] < 0
+		 || x[1] < 0
+		 || x[0] > height() - 1
+		 || x[1] > width() - 1)
+			return 0;
+		if (!x.defined())
+			return 0;
+		return 1;
+	}
+
 	pixel get_bl(point x) const {
 		pixel result[2];
 
@@ -335,35 +407,143 @@ public:
 	virtual image *scale_generator(int height, int width, int depth, char *name) const = 0;
 
 	/*
-	 * Return an image scaled by some factor >= 1.0
+	 * Generate an image of medians within a given radius
 	 */
-	image *scale(ale_pos f, char *name) const {
 
-		/*
-		 * Assertion:
-		 *
-		 * This is not a good way to scale down images, and we probably
-		 * don't want to scale images by a factor of 1.0.
-		 */
-		assert (f > 1.0);
+	image *medians(int radius) const {
 
-		image *is = scale_generator(
-			(int) floor(height() * f), 
-			(int) floor(width()  * f), depth(), name);
+		assert (radius >= 0);
+		
+		image *is = scale_generator(height(), width(), depth(), "median");
+		assert(is);
+
+		for (unsigned int i = 0; i < height(); i++)
+		for (unsigned int j = 0; j < width(); j++) {
+
+			std::vector<ale_real> p[3];
+
+			for (int ii = -radius; ii <= radius; ii++)
+			for (int jj = -radius; jj <= radius; jj++) {
+				int iii = i + ii;
+				int jjj = j + jj;
+
+				if (in_bounds(point(iii, jjj)))
+					for (int k = 0; k < 3; k++)
+						if (finite(get_pixel(iii, jjj)[k]))
+							p[k].push_back(get_pixel(iii, jjj)[k]);
+			}
+
+			is->pix(i, j) = d2::pixel::undefined();
+
+			for (int k = 0; k < 3; k++) {
+				std::sort(p[k].begin(), p[k].end());
+
+				unsigned int pkc = p[k].size();
+
+				if (pkc == 0)
+					continue;
+
+				if (pkc % 2 == 0) 
+					is->chan(i, j, k) = (p[k][pkc / 2] + p[k][pkc / 2 - 1]) / 2;
+				else
+					is->chan(i, j, k) = p[k][pkc / 2];
+			}
+		}
+
+		return is;
+	}
+
+	/*
+	 * Generate an image of differences of the first channel.  The first
+	 * coordinate differences are stored in the first channel, second in the
+	 * second channel.
+	 */
+
+	image *fcdiffs() const {
+		image *is = scale_generator(height(), width(), depth(), "diff");
 
 		assert(is);
 
-		unsigned int i, j, k;
+		for (unsigned int i = 0; i < height(); i++)
+		for (unsigned int j = 0; j < width(); j++) {
 
-		for (i = 0; i < is->height(); i++)
-		for (j = 0; j < is->width(); j++) 
-		for (k = 0; k < is->depth(); k++)
-			is->set_pixel(i, j,
-				get_scaled_bl(point(i, j), f));
+			if (i + 1 < height()
+			 && i > 0
+			 && !finite(chan(i, j, 0))) {
+
+				is->chan(i, j, 0) = (chan(i + 1, j, 0) - chan(i - 1, j, 0)) / 2;
+
+			} else if (i + 1 < height()
+			        && i > 0
+			        && finite(chan(i + 1, j, 0))
+			        && finite(chan(i - 1, j, 0))) {
+
+				is->chan(i, j, 0) = ((chan(i, j, 0) - chan(i - 1, j, 0))
+						   + (chan(i + 1, j, 0) - chan(i, j, 0))) / 2;
+
+			} else if (i + 1 < height()
+				&& finite(chan(i + 1, j, 0))) {
+
+				is->chan(i, j, 0) = chan(i + 1, j, 0) - chan(i, j, 0);
+
+			} else if (i > 0
+				&& finite(chan(i - 1, j, 0))) {
+				 
+				is->chan(i, j, 0) = chan(i, j, 0) - chan(i - 1, j, 0);
+
+			} else {
+				is->chan(i, j, 0) = 0;
+			}
+
+			if (j + 1 < width()
+			 && j > 0
+			 && !finite(chan(i, j, 0))) {
+
+				is->chan(i, j, 1) = (chan(i, j + 1, 0) - chan(i, j - 1, 0)) / 2;
+
+			} else if (j + 1 < width()
+			        && j > 0
+			        && finite(chan(i, j + 1, 0))
+			        && finite(chan(i, j - 1, 0))) {
+
+				is->chan(i, j, 1) = ((chan(i, j, 0) - chan(i, j - 1, 0))
+						   + (chan(i, j + 1, 0) - chan(i, j, 0))) / 2;
+
+			} else if (j + 1 < width() && finite(chan(i, j + 1, 0))) {
+
+				is->chan(i, j, 1) = chan(i, j + 1, 0) - chan(i, j, 0);
+
+			} else if (j > 0 && finite(chan(i, j - 1, 0))) {
+				 
+				is->chan(i, j, 1) = chan(i, j, 0) - chan(i, j - 1, 0);
+
+			} else {
+				is->chan(i, j, 1) = 0;
+			}
+		}
 
 		return is;
 	}
 
+	/*
+	 * Generate an image of median (within a given radius) difference of the
+	 * first channel.
+	 */
+
+	image *fcdiff_median(int radius) const {
+		image *diff = fcdiffs();
+
+		assert(diff);
+
+		image *median = diff->medians(radius);
+
+		assert(median);
+
+		delete diff;
+
+		return median;
+	}
+
 	/*
 	 * Scale by half.  We use the following filter:
 	 *
@@ -569,6 +749,47 @@ public:
 		return is;
 	}
 
+	/*
+	 * Return an image scaled by some factor != 1.0, using bilinear
+	 * interpolation.
+	 */
+	image *scale(ale_pos f, char *name) const {
+
+		/*
+		 * We probably don't want to scale images by a factor of 1.0,
+		 * or by non-positive values.
+		 */
+		assert (f != 1.0 && f > 0);
+
+		if (f > 1.0) {
+			image *is = scale_generator(
+				(int) floor(height() * f), 
+				(int) floor(width()  * f), depth(), name);
+
+			assert(is);
+
+			unsigned int i, j, k;
+
+			for (i = 0; i < is->height(); i++)
+			for (j = 0; j < is->width(); j++) 
+			for (k = 0; k < is->depth(); k++)
+				is->set_pixel(i, j,
+					get_scaled_bl(point(i, j), f));
+
+			is->_offset = point(_offset[0] * f, _offset[1] * f);
+
+			return is;
+		} else if (f == 0.5) {
+			return scale_by_half(name);
+		} else {
+			image *is = scale(2*f, name);
+			image *result = is->scale(0.5, name);
+			delete is;
+			return result;
+		}
+
+	}
+
 	/*
 	 * Scale an image definition array by 1/2.
 	 *
@@ -581,20 +802,10 @@ public:
 	 *
 	 * Whereas scaling of image weight arrays is not generally obvious in
 	 * either purpose or method, ALE requires that image definition arrays
-	 * be scalable, and the method we implement here is a fairly obvious
-	 * one.  In particular, if any source pixel contributing to the value of 
-	 * a scaled target pixel has an undefined value, then the scaled target
-	 * pixel is undefined (zero).  Otherwise, it is defined (non-zero).
-	 * 
-	 * Since there are many possible ways of implementing this function, we
-	 * choose an easy way and simply multiply the numerical values of the
-	 * source pixels to obtain the value of the scaled target pixel.
+	 * be scalable.
 	 *
-	 * XXX: we consider all pixels within a 9-pixel square to contribute.
-	 * There are other approaches.  For example, edge pixels could be
-	 * considered to have six contributing pixels and corner pixels four
-	 * contributing pixels.  To use this convention, change the ': 0' text
-	 * in the code below to ': 1'.
+	 * We currently use a geometric mean to implement scaling of
+	 * definition arrays.
 	 */
 
 	image *defined_scale_by_half(char *name) const {
@@ -607,35 +818,43 @@ public:
 		assert(is);
 
 		for (unsigned int i = 0; i < is->height(); i++)
-		for (unsigned int j = 0; j < is->width(); j++) 
+		for (unsigned int j = 0; j < is->width(); j++) {
 
-			is->set_pixel(i, j, 
+			pixel value = pixel
 
-				( ((i > 0 && j > 0) 
-				    ? get_pixel(2 * i - 1, 2 * j - 1)
-				    : pixel())
+			      ( ( ((i > 0 && j > 0) 
+				    ? ppow(get_pixel(2 * i - 1, 2 * j - 1), 0.0625) 
+				    : pixel(0, 0, 0))
 				* ((i > 0)
-				    ? get_pixel(2 * i - 1, 2 * j)
-				    : pixel())
+				    ? ppow(get_pixel(2 * i - 1, 2 * j), 0.125)
+				    : pixel(0, 0, 0))
 				* ((i > 0 && j < is->width() - 1)
-				    ? get_pixel(2 * i - 1, 2 * j + 1)
-				    : pixel())
+				    ? ppow(get_pixel(2 * i - 1, 2 * j + 1), 0.0625)
+				    : pixel(0, 0, 0))
 				* ((j > 0)
-				    ? get_pixel(2 * i, 2 * j - 1)
-				    : pixel())
-				* get_pixel(2 * i, 2 * j)
+				    ? ppow(get_pixel(2 * i, 2 * j - 1), 0.125)
+				    : pixel(0, 0, 0))
+				* ppow(get_pixel(2 * i, 2 * j), 0.25) 
 				* ((j < is->width() - 1)
-				    ? get_pixel(2 * i, 2 * j + 1)
-				    : pixel())
+				    ? ppow(get_pixel(2 * i, 2 * j + 1), 0.125)
+				    : pixel(0, 0, 0))
 				* ((i < is->height() - 1 && j > 0)
-				    ? get_pixel(2 * i + 1, 2 * j - 1)
-				    : pixel())
+				    ? ppow(get_pixel(2 * i + 1, 2 * j - 1), 0.0625)
+				    : pixel(0, 0, 0))
 				* ((i < is->height() - 1)
-				    ? get_pixel(2 * i + 1, 2 * j)
-				    : pixel())
+				    ? ppow(get_pixel(2 * i + 1, 2 * j), 0.125)
+				    : pixel(0, 0, 0))
 				* ((i < is->height() && j < is->width() - 1)
-				    ? get_pixel(2 * i + 1, 2 * j + 1)
-				    : pixel())));
+				    ? ppow(get_pixel(2 * i + 1, 2 * j + 1), 0.0625)
+				    : pixel(0, 0, 0))));
+
+
+			for (int k = 0; k < 3; k++)
+				if (!finite(value[k]))
+					value[k] = 0;
+
+			is->set_pixel(i, j, value);
+		}
 
 		is->_offset = point(_offset[0] * f, _offset[1] * f);
 
@@ -644,10 +863,10 @@ public:
 
 	/*
 	 * Extend the image area to the top, bottom, left, and right,
-	 * initializing the new image areas with black pixels.
+	 * initializing the new image areas with black pixels.  Negative values
+	 * shrink the image.
 	 */
-	virtual void extend(unsigned int top, unsigned int bottom, 
-			unsigned int left, unsigned int right) = 0;
+	virtual void extend(int top, int bottom, int left, int right) = 0;
 
 	/*
 	 * Clone 
