@@ -25,10 +25,10 @@
  */
 
 unsigned int render::rx_count;
-int *render::rx_parameters;
+exclusion *render::rx_parameters;
 int render::rx_show;
 render *render::directory[ACTIVE_RENDERER_COUNT];
 int render::directory_length;
 int render::extend;
 double render::scale_factor;
-double render::wt = 0.8;
+double render::wt = 0.0001;
