@@ -115,6 +115,8 @@ protected:
         unsigned int iterations;
 	psf *lresponse, *nlresponse;
 	int exposure_register;
+	int use_weighted_median;
+	double weight_limit;
 
 	/*
 	 * Calculate the simulated input frame SIMULATED from the image
@@ -122,26 +124,50 @@ protected:
 	 * pixels.  
 	 */
 
-	void _ip_frame_simulate(image *approximation, 
-			image *lsimulated, image *nlsimulated, 
-			transformation t, const raster *lresponse, 
-			const raster *nlresponse, const exposure &exp) {
+	struct sim_subdomain_args {
+		int frame_num;
+		image *approximation;
+		image *lsimulated;
+		image *nlsimulated;
+		image *lsim_weights;
+		image *nlsim_weights;
+		transformation t;
+		const raster *lresponse;
+		const raster *nlresponse;
+		const exposure *exp;
+		int i_min, i_max, j_min, j_max;
+#ifdef USE_PTHREAD
+		pthread_mutex_t *lock;
+#endif
+	};
 
-		/*
-		 * Initializations for linear filtering
-		 */
+	static void *_ip_frame_simulate_subdomain_linear(void *args) {
 
-		image_ale_real *lsim_weights = new image_ale_real(
-				lsimulated->height(),
-				lsimulated->width(),
-				lsimulated->depth());
+		sim_subdomain_args *sargs = (sim_subdomain_args *) args;
+
+		int frame_num = sargs->frame_num;
+		image *approximation = sargs->approximation;
+		image *lsimulated = sargs->lsimulated;
+		image *lsim_weights = sargs->lsim_weights;
+		transformation t = sargs->t;
+		const raster *lresponse = sargs->lresponse;
+		unsigned int i_min = sargs->i_min;
+		unsigned int i_max = sargs->i_max;
+		unsigned int j_min = sargs->j_min;
+		unsigned int j_max = sargs->j_max;
+#ifdef USE_PTHREAD
+		pthread_mutex_t *lock = sargs->lock;
+#endif
 
 		/*
 		 * Linear filtering, iterating over approximation pixels
 		 */
 
-                for (unsigned int i = 0; i < approximation->height(); i++)
-                for (unsigned int j = 0; j < approximation->width();  j++) {
+                for (unsigned int i = i_min; i < i_max; i++)
+                for (unsigned int j = j_min; j < j_max; j++) {
+
+			if (is_excluded_r(approximation->offset(), i, j, frame_num))
+				continue;
 
 			/*
 			 * Obtain the position Q and dimensions D of
@@ -153,6 +179,13 @@ protected:
 			point q;
 			ale_pos d[2];
 
+			/*
+			 * XXX: This appears to calculate the wrong thing.
+			 */
+
+			if (is_excluded_f(p, frame_num))
+				continue;
+
 			t.unscaled_map_area_inverse(p, &q, d);
 			
 			/*
@@ -187,59 +220,46 @@ protected:
 						    lef - jj, rig - jj,
 						    lresponse->select(ii, jj));
 
+#ifdef USE_PTHREAD
+				pthread_mutex_lock(lock);
+#endif
 				lsimulated->pix(ii, jj) +=
 					r(approximation->get_pixel(i, j));
 				lsim_weights->pix(ii, jj) +=
 					r.weight();
+#ifdef USE_PTHREAD
+				pthread_mutex_unlock(lock);
+#endif
                         }
                 }
 
-		/*
-		 * Normalize linear
-		 */
+		return NULL;
+	}
 
-		for (unsigned int ii = 0; ii < lsimulated->height(); ii++)
-		for (unsigned int jj = 0; jj < lsimulated->width(); jj++) {
-			pixel weight = lsim_weights->get_pixel(ii, jj);
+	static void *_ip_frame_simulate_subdomain_nonlinear(void *args) {
 
-			if (weight[0] > 0.1
-			 && weight[1] > 0.1 
-			 && weight[2] > 0.1)
-				lsimulated->pix(ii, jj)
-					/= weight;
-			else
-				lsimulated->pix(ii, jj)
-					/= 0;  /* Generate a non-finite value */
-		}
+		sim_subdomain_args *sargs = (sim_subdomain_args *) args;
 
-		/*
-		 * Finalize linear
-		 */
-
-		delete lsim_weights;
-
-		/*
-		 * Return if there is no non-linear step.
-		 */
-
-		if (nlsimulated == NULL)
-			return;
-
-		/*
-		 * Initialize non-linear
-		 */
-
-		image_ale_real *nlsim_weights = new image_ale_real(
-				nlsimulated->height(),
-				nlsimulated->width(),
-				nlsimulated->depth());
+		image *lsimulated = sargs->lsimulated;
+		image *nlsimulated = sargs->nlsimulated;
+		image *nlsim_weights = sargs->nlsim_weights;
+		transformation t = sargs->t;
+		const raster *nlresponse = sargs->nlresponse;
+		const exposure *exp = sargs->exp;
+		unsigned int i_min = sargs->i_min;
+		unsigned int i_max = sargs->i_max;
+		unsigned int j_min = sargs->j_min;
+		unsigned int j_max = sargs->j_max;
+#ifdef USE_PTHREAD
+		pthread_mutex_t *lock = sargs->lock;
+#endif
 
 		/*
 		 * Iterate non-linear
 		 */
 
-                for (unsigned int i = 0; i < lsimulated->height(); i++)
-                for (unsigned int j = 0; j < lsimulated->width();  j++) {
+                for (unsigned int i = i_min; i < i_max; i++)
+                for (unsigned int j = j_min; j < j_max; j++) {
 
 			/*
 			 * Convenient variables for expressing the boundaries
@@ -273,13 +293,153 @@ protected:
 						    lef - jj, rig - jj,
 						    nlresponse->select(ii, jj));
 
+#ifdef USE_PTHREAD
+				pthread_mutex_lock(lock);
+#endif
 				nlsimulated->pix(ii, jj) +=
-					r(exp.unlinearize(lsimulated->get_pixel(i, j)));
+					r(exp->unlinearize(lsimulated->get_pixel(i, j)));
 				nlsim_weights->pix(ii, jj) +=
 					r.weight();
+#ifdef USE_PTHREAD
+				pthread_mutex_unlock(lock);
+#endif
                         }
                 }
 
+		return NULL;
+	}
+
+	void _ip_frame_simulate(int frame_num, image *approximation, 
+			image *lsimulated, image *nlsimulated, 
+			transformation t, const raster *lresponse, 
+			const raster *nlresponse, const exposure &exp) {
+
+		/*
+		 * Threading initializations
+		 */
+
+		int N;
+
+#ifdef USE_PTHREAD
+		N = thread::count();
+
+		pthread_t *threads = (pthread_t *) malloc(sizeof(pthread_t) * N);
+		pthread_attr_t *thread_attr = (pthread_attr_t *) malloc(sizeof(pthread_attr_t) * N);
+		pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
+
+#else
+		N = 1;
+#endif
+
+		sim_subdomain_args *args = (sim_subdomain_args *) malloc(sizeof(sim_subdomain_args) * N);
+
+
+		/*
+		 * Initializations for linear filtering
+		 */
+
+		image_ale_real *lsim_weights = new image_ale_real(
+				lsimulated->height(),
+				lsimulated->width(),
+				lsimulated->depth());
+
+
+		for (int ti = 0; ti < N; ti++) {
+			args[ti].frame_num = frame_num;
+			args[ti].approximation = approximation;
+			args[ti].lsimulated = lsimulated;
+			args[ti].nlsimulated = nlsimulated;
+			args[ti].lsim_weights = lsim_weights;
+			args[ti].t = t;
+			args[ti].lresponse = lresponse;
+			args[ti].nlresponse = nlresponse;
+			args[ti].exp = &exp;
+			args[ti].i_min = (approximation->height() * ti) / N;
+			args[ti].i_max = (approximation->height() * (ti + 1)) / N;
+			args[ti].j_min = 0;
+			args[ti].j_max = approximation->width();
+
+#ifdef USE_PTHREAD
+			args[ti].lock = &lock;
+			pthread_attr_init(&thread_attr[ti]);
+			pthread_attr_setdetachstate(&thread_attr[ti], PTHREAD_CREATE_JOINABLE);
+			pthread_create(&threads[ti], &thread_attr[ti], _ip_frame_simulate_subdomain_linear, &args[ti]);
+#else
+			_ip_frame_simulate_subdomain_linear(&args[ti]);
+#endif
+		}
+
+#ifdef USE_PTHREAD
+		for (int ti = 0; ti < N; ti++) {
+			pthread_join(threads[ti], NULL);
+		}
+#endif
+
+		/*
+		 * Normalize linear
+		 */
+
+		for (unsigned int ii = 0; ii < lsimulated->height(); ii++)
+		for (unsigned int jj = 0; jj < lsimulated->width(); jj++) {
+			pixel weight = lsim_weights->get_pixel(ii, jj);
+			const ale_real weight_floor = 0.00001;
+
+			if (weight[0] > weight_floor
+			 && weight[1] > weight_floor
+			 && weight[2] > weight_floor)
+				lsimulated->pix(ii, jj)
+					/= weight;
+			else
+				lsimulated->pix(ii, jj)
+					/= 0;  /* Generate a non-finite value */
+		}
+
+		/*
+		 * Finalize linear
+		 */
+
+		delete lsim_weights;
+
+		/*
+		 * Return if there is no non-linear step.
+		 */
+
+		if (nlsimulated == NULL) {
+			delete args;
+			return;
+		}
+
+		/*
+		 * Initialize non-linear
+		 */
+
+		image_ale_real *nlsim_weights = new image_ale_real(
+				nlsimulated->height(),
+				nlsimulated->width(),
+				nlsimulated->depth());
+
+		for (int ti = 0; ti < N; ti++) {
+			args[ti].nlsim_weights = nlsim_weights;
+			args[ti].i_min = (lsimulated->height() * ti) / N;
+			args[ti].i_max = (lsimulated->height() * (ti + 1)) / N;
+			args[ti].j_min = 0;
+			args[ti].j_max = lsimulated->width();
+
+#ifdef USE_PTHREAD
+			pthread_attr_init(&thread_attr[ti]);
+			pthread_attr_setdetachstate(&thread_attr[ti], PTHREAD_CREATE_JOINABLE);
+			pthread_create(&threads[ti], &thread_attr[ti], _ip_frame_simulate_subdomain_nonlinear, &args[ti]);
+#else
+			_ip_frame_simulate_subdomain_nonlinear(&args[ti]);
+#endif
+		}
+
+#ifdef USE_PTHREAD
+		for (int ti = 0; ti < N; ti++) {
+			pthread_join(threads[ti], NULL);
+		}
+#endif
+
 		/*
 		 * Normalize non-linear
 		 */
@@ -287,10 +447,11 @@ protected:
 		for (unsigned int ii = 0; ii < nlsimulated->height(); ii++)
 		for (unsigned int jj = 0; jj < nlsimulated->width(); jj++) {
 			pixel weight = nlsim_weights->get_pixel(ii, jj);
+			ale_real weight_floor = 0.00001;
 
-			if (weight[0] > 0.1
-			 && weight[1] > 0.1 
-			 && weight[2] > 0.1)
+			if (weight[0] > weight_floor
+			 && weight[1] > weight_floor
+			 && weight[2] > weight_floor)
 				nlsimulated->pix(ii, jj)
 					/= weight;
 			else
@@ -303,9 +464,122 @@ protected:
 		 */
 
 		delete nlsim_weights;
+
+		delete args;
 	}
 
+	struct correction_t {
+		/*
+		 * Type
+		 */
+		int use_weighted_median;
 
+		/*
+		 * Weighted Median
+		 */
+		image_weighted_median *iwm;
+
+		/*
+		 * Weight limit
+		 */
+		double weight_limit;
+
+		/*
+		 * Common
+		 */
+		image *c;
+		image *cc;
+
+		/*
+		 * Create correction structures.
+		 */
+		correction_t(int use_weighted_median, double weight_limit, unsigned int h, unsigned int w, unsigned int d) {
+			this->use_weighted_median = use_weighted_median;
+			if (use_weighted_median)
+				iwm = new image_weighted_median(h, w, d);
+			this->weight_limit = weight_limit;
+			c = new image_ale_real(h, w, d);
+			cc = new image_ale_real(h, w, d);
+		}
+
+		/*
+		 * Destroy correction structures
+		 */
+		~correction_t() {
+			if (use_weighted_median)
+				delete iwm;
+			delete c;
+			delete cc;
+		}
+
+		/*
+		 * Correction count
+		 */
+		pixel get_count(int i, int j) {
+			if (use_weighted_median)
+				return iwm->get_weights()->get_pixel(i, j);
+			else
+				return cc->get_pixel(i, j);
+		}
+
+		/*
+		 * Check for weight limit
+		 */
+		int weight_limited(int i, int j) {
+			if (weight_limit
+			 && get_count(i, j)[0] >= weight_limit
+			 && get_count(i, j)[1] >= weight_limit
+			 && get_count(i, j)[2] >= weight_limit)
+				return 1;
+
+			return 0;
+		}
+
+		/*
+		 * Correction value
+		 */
+		pixel get_correction(int i, int j) {
+			if (use_weighted_median)
+				return iwm->get_colors()->get_pixel(i, j);
+			else
+				return c->get_pixel(i, j)
+				     / cc->get_pixel(i, j);
+		}
+
+		/*
+		 * Frame end
+		 */
+		void frame_end(int frame_num) {
+			if (use_weighted_median) {
+
+				for (unsigned int i = 0; i < c->height(); i++)
+				for (unsigned int j = 0; j < c->width(); j++) {
+
+					/*
+					 * Update the median calculator
+					 */
+					pixel cval = c->get_pixel(i, j);
+					pixel ccval = cc->get_pixel(i, j);
+					iwm->accumulate(i, j, frame_num, cval / ccval, ccval);
+
+					/*
+					 * Reset the counters
+					 */
+					c->pix(i, j) = pixel::zero();
+					cc->pix(i, j) = pixel::zero();
+				}
+			}
+		}
+
+		/*
+		 * Update correction structures, using either a weighted mean update or
+		 * a weighted median update.
+		 */
+		void update(int i, int j, pixel value_times_weight, pixel weight) {
+			c->pix(i, j) += value_times_weight;
+			cc->pix(i, j) += weight;
+		}
+	};
 
 	/*
 	 * For each pixel in APPROXIMATION, calculate the differences
@@ -325,163 +599,163 @@ protected:
 	 * densities.
 	 */
 
-	void _ip_frame_correct(image *approximation, image *c, 
-			image *cc, const image *real, image *lsimulated, 
-			image *nlsimulated, transformation t, 
-			const backprojector *lresponse, 
-			const backprojector *nlresponse) {
+	struct correct_subdomain_args {
+		int frame_num;
+		image *approximation;
+		correction_t *cu;
+		const image *real;
+		image *lreal;
+		image *lsimulated;
+		image *nlsimulated;
+		transformation t;
+		const backprojector *lresponse;
+		const backprojector *nlresponse;
+		unsigned int i_min, i_max, j_min, j_max;
+		double weight_limit;
+	};
+
+	static void *_ip_frame_correct_subdomain_nonlinear(void *args) {
+
+		correct_subdomain_args *sargs = (correct_subdomain_args *) args;
+
+		int frame_num = sargs->frame_num;
+		image *approximation = sargs->approximation;
+		const image *real = sargs->real;
+		image *lreal = sargs->lreal;
+		image *lsimulated = sargs->lsimulated;
+		const image *nlsimulated = sargs->nlsimulated;
+		transformation t = sargs->t;
+		const backprojector *nlresponse = sargs->nlresponse;
+		unsigned int i_min = sargs->i_min;
+		unsigned int i_max = sargs->i_max;
+		unsigned int j_min = sargs->j_min;
+		unsigned int j_max = sargs->j_max;
 
 		/*
-		 * Generate the image to compare lsimulated with.
+		 * Unlinearize values from lsimulated.
 		 */
 
-		const image *lreal;
+		for (unsigned int i = i_min; i < i_max; i++)
+		for (unsigned int j = j_min; j < j_max; j++)
+			lreal->set_pixel(i, j, real->exp().unlinearize(
+						lsimulated->get_pixel(i, j)));
 
-		if (nlsimulated == NULL)
-			lreal = real;
-		else {
 
-			image *new_lreal = new image_ale_real(
-					real->height(),
-					real->width(),
-					real->depth(),
-					"IPC lreal", 
-					&real->exp());
+		/*
+		 * Backproject from real to lreal, iterating over all pixels
+		 * in lreal.
+		 */
+
+		for (unsigned int i = i_min; i < i_max; i++)
+		for (unsigned int j = j_min; j < j_max; j++) {
 
 			/*
-			 * Unlinearize values from lsimulated.
+			 * XXX: Is this right?
 			 */
 
-			for (unsigned int i = 0; i < lsimulated->height(); i++)
-			for (unsigned int j = 0; j < lsimulated->width(); j++)
-				new_lreal->set_pixel(i, j, real->exp().unlinearize(
-							lsimulated->get_pixel(i, j)));
+			if (is_excluded_r(approximation->offset(), i, j, frame_num))
+				continue;
 
 			/*
-			 * Backproject from real to lreal, iterating over all pixels
-			 * in lreal.
+			 * Convenient variables for expressing the boundaries
+			 * of the mapped area.
+			 */
+			
+			ale_pos top = i - 0.5;
+			ale_pos bot = i + 0.5;
+			ale_pos lef = j - 0.5;
+			ale_pos rig = j + 0.5;
+
+			/*
+			 * Iterate over non-linear pixels influenced by linear
+			 * pixels.
 			 */
 
-			for (unsigned int i = 0; i < new_lreal->height(); i++)
-			for (unsigned int j = 0; j < new_lreal->width();  j++) {
+			for (int ii = (int) floor(top + nlresponse->min_i());
+				ii <= ceil(bot + nlresponse->max_i()); ii++)
+			for (int jj = (int) floor(lef + nlresponse->min_j());
+				jj <= ceil(rig + nlresponse->max_j()); jj++) {
+
+				if (ii < (int) 0
+				 || ii >= (int) nlsimulated->height()
+				 || jj < (int) 0
+				 || jj >= (int) nlsimulated->width())
+					continue;
+
+				class backprojector::psf_result r =
+					(*nlresponse)(top - ii, bot - ii,
+						    lef - jj, rig - jj,
+						    nlresponse->select(ii, jj));
+
+
+				pixel comp_real =
+					real->exp().unlinearize(real->get_pixel(ii, jj));
+
+				pixel comp_simu =
+					nlsimulated->get_pixel(ii, jj);
+
+				if (!finite(comp_simu[0])
+				 || !finite(comp_simu[1])
+				 || !finite(comp_simu[2]))
+					continue;
 
 				/*
-				 * Convenient variables for expressing the boundaries
-				 * of the mapped area.
+				 * Backprojection value.
 				 */
-				
-				ale_pos top = i - 0.5;
-				ale_pos bot = i + 0.5;
-				ale_pos lef = j - 0.5;
-				ale_pos rig = j + 0.5;
+
+				pixel bpv = r(comp_real - comp_simu);
 
 				/*
-				 * Iterate over non-linear pixels influenced by linear
-				 * pixels.
+				 * Error calculation
 				 */
 
-				for (int ii = (int) floor(top + nlresponse->min_i());
-					ii <= ceil(bot + nlresponse->max_i()); ii++)
-				for (int jj = (int) floor(lef + nlresponse->min_j());
-					jj <= ceil(rig + nlresponse->max_j()); jj++) {
-
-					if (ii < (int) 0
-					 || ii >= (int) nlsimulated->height()
-					 || jj < (int) 0
-					 || jj >= (int) nlsimulated->width())
-						continue;
-
-					class backprojector::psf_result r =
-						(*nlresponse)(top - ii, bot - ii,
-							    lef - jj, rig - jj,
-							    nlresponse->select(ii, jj));
-
-
-					pixel comp_real =
-						real->exp().unlinearize(real->get_pixel(ii, jj));
-
-					pixel comp_simu =
-						nlsimulated->get_pixel(ii, jj);
-
-					if (!finite(comp_simu[0])
-					 || !finite(comp_simu[1])
-					 || !finite(comp_simu[2]))
-						continue;
-
-					/*
-					 * Backprojection value.
-					 */
-
-					pixel bpv = r(comp_real - comp_simu);
-
-					/*
-					 * Error calculation
-					 */
-
-					new_lreal->pix(i, j) += bpv;
-				}
+				lreal->pix(i, j) += bpv;
 			}
-
-			/*
-			 * Linearize lreal.
-			 */
-
-			for (unsigned int i = 0; i < new_lreal->height(); i++)
-			for (unsigned int j = 0; j < new_lreal->width(); j++)
-				new_lreal->set_pixel(i, j, real->exp().linearize(
-							new_lreal->get_pixel(i, j)));
-
-			/*
-			 * Perform exposure adjustment.
-			 */
-
-			if (exposure_register) {
-
-				pixel ec;
-
-#if 0
-				ec = lsimulated->avg_channel_magnitude()
-				   / new_lreal->avg_channel_magnitude();
-#else
-
-				pixel_accum ratio_sum;
-				pixel_accum weight_sum;
-
-				for (unsigned int i = 0; i < new_lreal->height(); i++)
-				for (unsigned int j = 0; j < new_lreal->width(); j++) {
-					pixel s = lsimulated->get_pixel(i, j);
-					pixel r = new_lreal->get_pixel(i, j);
-					pixel confidence = real->exp().confidence(r);
-
-					if (s[0] > 0.001 
-					 && s[1] > 0.001
-					 && s[2] > 0.001
-					 && r[0] > 0.001
-					 && r[1] > 0.001
-					 && r[2] > 0.001) {
-						ratio_sum += confidence * s / r;
-						weight_sum += confidence;
-					}
-				}
-
-				ec = ratio_sum / weight_sum;
-
-#endif
-
-				real->exp().set_multiplier(
-					real->exp().get_multiplier() * ec);
-			}
-
-			lreal = new_lreal;
 		}
 
+		/*
+		 * Linearize lreal.
+		 */
+
+		for (unsigned int i = i_min; i < i_max; i++)
+		for (unsigned int j = j_min; j < j_max; j++)
+			lreal->set_pixel(i, j, real->exp().linearize(
+						lreal->get_pixel(i, j)));
+
+
+		return NULL;
+	}
+
+	static void *_ip_frame_correct_subdomain_linear(void *args) {
+
+		correct_subdomain_args *sargs = (correct_subdomain_args *) args;
+
+		int frame_num = sargs->frame_num;
+		image *approximation = sargs->approximation;
+		correction_t *cu = sargs->cu;
+		const image *real = sargs->real;
+		const image *lreal = sargs->lreal;
+		const image *lsimulated = sargs->lsimulated;
+		transformation t = sargs->t;
+		const backprojector *lresponse = sargs->lresponse;
+		unsigned int i_min = sargs->i_min;
+		unsigned int i_max = sargs->i_max;
+		unsigned int j_min = sargs->j_min;
+		unsigned int j_max = sargs->j_max;
 
 		/*
 		 * Iterate over all pixels in the approximation.
 		 */
 
-                for (unsigned int i = 0; i < approximation->height(); i++)
-                for (unsigned int j = 0; j < approximation->width();  j++) {
+                for (unsigned int i = i_min; i < i_max; i++)
+                for (unsigned int j = j_min; j < j_max; j++) {
+
+			/*
+			 * Check correction count against any weight limit.
+			 */
+
+			if (cu->weight_limited(i, j))
+				continue;
 
 			/*
 			 * Obtain the position Q and dimensions D of image
@@ -521,6 +795,9 @@ protected:
 				 || jj >= (int) lreal->width())
 					continue;
 
+				if (is_excluded_f(ii, jj, frame_num))
+					continue;
+
 				unsigned int selection = lresponse->select(ii, jj);
 
 				class backprojector::psf_result r =
@@ -568,8 +845,21 @@ protected:
 				 * confidence is uniform].
 				 */
 
-				pixel conf = real->exp().one_sided_confidence(comp_lreal, bpv);
+				// Ordinary certainty
+				// pixel conf = real->exp().confidence(comp_lreal);
+
+				// One-sided certainty
+				// pixel conf = real->exp().one_sided_confidence(comp_lreal, bpv);
 				      // conf = real->exp().one_sided_confidence(comp_real, bpv);
+				      
+				// Estimate-based certainty
+				// pixel conf = real->exp().confidence(comp_simu);
+					
+				// One-sided estimate-based certainty
+				pixel conf = real->exp().one_sided_confidence(comp_simu, bpv);
+
+				// One-sided approximation-based certainty
+				// pixel conf = real->exp().one_sided_confidence(approximation->pix(i, j), bpv);
 					
 				/*
 				 * If a color is bayer-interpolated, then we have no confidence in its
@@ -586,7 +876,7 @@ protected:
 				 * Error calculation
 				 */
 
-				c->pix(i, j) += bpv * conf;
+				// c->pix(i, j) += bpv * conf;
 
 				/*
 				 * Increment the backprojection weight.  When
@@ -594,11 +884,180 @@ protected:
 				 * each frame's correction equally.
 				 */
 
-				cc->pix(i, j) += conf * r.weight() 
-					       / lresponse->integral(selection);
+				// cc->pix(i, j) += conf * r.weight() 
+				//	       / lresponse->integral(selection);
+
+				cu->update(i, j, bpv * conf, conf * r.weight() / lresponse->integral(selection));
 			}
                 }
 
+		return NULL;
+	}
+
+	void _ip_frame_correct(int frame_num, image *approximation,
+			correction_t *cu, const image *real, image *lsimulated, 
+			image *nlsimulated, transformation t, 
+			const backprojector *lresponse, 
+			const backprojector *nlresponse) {
+
+		/*
+		 * Threading initializations
+		 */
+
+		int N;
+
+#ifdef USE_PTHREAD
+		N = thread::count();
+
+		pthread_t *threads = (pthread_t *) malloc(sizeof(pthread_t) * N);
+		pthread_attr_t *thread_attr = (pthread_attr_t *) malloc(sizeof(pthread_attr_t) * N);
+
+#else
+		N = 1;
+#endif
+
+		correct_subdomain_args *args = (correct_subdomain_args *) malloc(sizeof(correct_subdomain_args) * N);
+
+		for (int ti = 0; ti < N; ti++) {
+			args[ti].frame_num = frame_num;
+			args[ti].approximation = approximation;
+			args[ti].cu = cu;
+			args[ti].real = real;
+			args[ti].lsimulated = lsimulated;
+			args[ti].nlsimulated = nlsimulated;
+			args[ti].t = t;
+			args[ti].lresponse = lresponse;
+			args[ti].nlresponse = nlresponse;
+		}
+
+		/*
+		 * Generate the image to compare lsimulated with.
+		 */
+
+		const image *lreal;
+
+		if (nlsimulated == NULL)
+			lreal = real;
+		else {
+
+			image *new_lreal = new image_ale_real(
+					real->height(),
+					real->width(),
+					real->depth(),
+					"IPC lreal", 
+					&real->exp());
+
+			for (int ti = 0; ti < N; ti++) {
+				args[ti].lreal = new_lreal;
+				args[ti].i_min = (new_lreal->height() * ti) / N;
+				args[ti].i_max = (new_lreal->height() * (ti + 1)) / N;
+				args[ti].j_min = 0;
+				args[ti].j_max = new_lreal->width();
+
+#ifdef USE_PTHREAD
+				pthread_attr_init(&thread_attr[ti]);
+				pthread_attr_setdetachstate(&thread_attr[ti], PTHREAD_CREATE_JOINABLE);
+				pthread_create(&threads[ti], &thread_attr[ti], _ip_frame_correct_subdomain_nonlinear, &args[ti]);
+#else
+				_ip_frame_correct_subdomain_nonlinear(&args[ti]);
+#endif
+			}
+
+#ifdef USE_PTHREAD
+			for (int ti = 0; ti < N; ti++) {
+				pthread_join(threads[ti], NULL);
+			}
+#endif
+
+			lreal = new_lreal;
+		}
+
+		/*
+		 * Perform exposure adjustment.
+		 */
+
+		if (exposure_register) {
+
+			pixel ec;
+
+#if 0
+			ec = lsimulated->avg_channel_magnitude()
+			   / lreal->avg_channel_magnitude();
+#elsif 0
+			pixel_accum ratio_sum;
+			pixel_accum weight_sum;
+
+			for (unsigned int i = 0; i < lreal->height(); i++)
+			for (unsigned int j = 0; j < lreal->width(); j++) {
+				pixel s = lsimulated->get_pixel(i, j);
+				pixel r = lreal->get_pixel(i, j);
+				pixel confidence = real->exp().confidence(r);
+
+				if (s[0] > 0.001 
+				 && s[1] > 0.001
+				 && s[2] > 0.001
+				 && r[0] > 0.001
+				 && r[1] > 0.001
+				 && r[2] > 0.001) {
+					ratio_sum += confidence * s / r;
+					weight_sum += confidence;
+				}
+			}
+
+			ec = ratio_sum / weight_sum;
+#else
+			pixel_accum ssum, rsum;
+
+			for (unsigned int i = 0; i < lreal->height(); i++)
+			for (unsigned int j = 0; j < lreal->width(); j++) {
+				pixel s = lsimulated->get_pixel(i, j);
+				pixel r = lreal->get_pixel(i, j);
+
+				pixel confidence = real->exp().confidence(s);
+
+				if (!s.finite()
+				 || !r.finite()
+				 || !confidence.finite())
+					continue;
+
+				ssum += confidence * s;
+				rsum += confidence * r;
+
+			}
+
+			ec = ssum / rsum;
+
+#endif
+
+			if (ec.finite() 
+			 && rsum[0] > 0.001
+			 && rsum[1] > 0.001
+			 && rsum[2] > 0.001)
+				real->exp().set_multiplier(
+					real->exp().get_multiplier() * ec);
+		}
+		for (int ti = 0; ti < N; ti++) {
+			args[ti].lreal = (d2::image *) lreal;
+			args[ti].i_min = (approximation->height() * ti) / N;
+			args[ti].i_max = (approximation->height() * (ti + 1)) / N;
+			args[ti].j_min = 0;
+			args[ti].j_max = approximation->width();
+
+#ifdef USE_PTHREAD
+			pthread_attr_init(&thread_attr[ti]);
+			pthread_attr_setdetachstate(&thread_attr[ti], PTHREAD_CREATE_JOINABLE);
+			pthread_create(&threads[ti], &thread_attr[ti], _ip_frame_correct_subdomain_linear, &args[ti]);
+#else
+			_ip_frame_correct_subdomain_linear(&args[ti]);
+#endif
+		}
+
+#ifdef USE_PTHREAD
+		for (int ti = 0; ti < N; ti++) {
+			pthread_join(threads[ti], NULL);
+		}
+#endif
+
 		if (nlsimulated)
 			delete lreal;
 	}
@@ -609,10 +1068,12 @@ protected:
 	 * count CC for affected pixels in C.
 	 */
 
-	virtual void _ip_frame(image *c, image *cc, const image *real, 
+	virtual void _ip_frame(int frame_num, correction_t *cu, const image *real, 
 			transformation t, const raster *f, const backprojector *b,
 			const raster *nlf, const backprojector *nlb) {
 
+		ui::get()->d2_irani_peleg_start();
+
 		/*
 		 * Initialize simulated data structures
 		 */
@@ -634,13 +1095,15 @@ protected:
 		 * Create simulated frames with forward projection.
 		 */
 
-		_ip_frame_simulate(approximation, lsimulated, nlsimulated, t, f, nlf, real->exp());
+		ui::get()->ip_frame_simulate_start();
+		_ip_frame_simulate(frame_num, approximation, lsimulated, nlsimulated, t, f, nlf, real->exp());
 
 		/*
 		 * Update the correction array using backprojection.
 		 */
 
-		_ip_frame_correct(approximation, c, cc, real, lsimulated, nlsimulated, t, b, nlb);
+		ui::get()->ip_frame_correct_start();
+		_ip_frame_correct(frame_num, approximation, cu, real, lsimulated, nlsimulated, t, b, nlb);
 
 		/*
 		 * Finalize data structures.
@@ -648,6 +1111,8 @@ protected:
 
                 delete lsimulated;
 		delete nlsimulated;
+
+		ui::get()->d2_irani_peleg_stop();
         }
 
 	/*
@@ -686,12 +1151,9 @@ protected:
 
                 for (unsigned int n = 0; n < iterations; n++) {
 
-                        image *correction = new image_ale_real (
-                                        approximation->height(),
-                                        approximation->width(),
-                                        approximation->depth());
-
-			image *correction_count = new image_ale_real(
+			correction_t *correction = new correction_t(
+					use_weighted_median,
+					weight_limit,
 					approximation->height(),
 					approximation->width(),
 					approximation->depth());
@@ -705,66 +1167,67 @@ protected:
 				if (!align::match(m))
 					continue;
 
+				ui::get()->ip_frame_start(m);
+
 				transformation t = align::of(m);
 				const image *real = image_rw::open(m);
 
-                                _ip_frame(correction, correction_count, real,
+                                _ip_frame(m, correction, real,
 					t, f[m], b[m], nlf, nlb);
 
 				image_rw::close(m);
+
+				correction->frame_end(m);
 			}
 
 			/*
 			 * Update the approximation.
 			 */
 
+			ui::get()->ip_update();
+
 			for (unsigned int i = 0; i < approximation->height(); i++)
 			for (unsigned int j = 0; j < approximation->width();  j++) {
 
-				pixel  cpix = correction       ->get_pixel(i, j);
-				pixel ccpix = correction_count ->get_pixel(i, j);
-				pixel  apix = approximation    ->get_pixel(i, j);
+				pixel  cpix = correction->get_correction(i, j);
+				pixel ccpix = correction->get_count(i, j);
+				pixel  apix = approximation->get_pixel(i, j);
 
-				if (ccpix[0] < 0.01
-				 || ccpix[1] < 0.01
-				 || ccpix[2] < 0.01)
-					continue;
+				for (unsigned int k = 0; k < 3; k++) {
 
-				pixel new_value = cpix / ccpix + apix;
+					const ale_real cc_floor = 0.0000001;
 
-				assert (finite(apix[0])
-				     && finite(apix[1])
-				     && finite(apix[2]));
+					if (ccpix[k] < cc_floor)
+						continue;
 
-				assert (finite(ccpix[0])
-				     && finite(ccpix[1])
-				     && finite(ccpix[2]));
+					if (!finite(cpix[k]))
+						continue;
 
-				assert (finite(cpix[0])
-				     && finite(cpix[1])
-				     && finite(cpix[2]));
+					ale_real new_value = cpix[k] + apix[k];
 
-				assert (finite(new_value[0])
-				     && finite(new_value[1])
-				     && finite(new_value[2]));
+					assert (finite(apix[k]));
+					assert (finite(ccpix[k]));
+					assert (finite(cpix[k]));
+					assert (finite(new_value));
 
-				/*
-				 * Negative light doesn't make sense.
-				 */
-				for (unsigned int k = 0; k < 3; k++)
-					if (new_value[k] < 0)
-						new_value[k] = 0;
+					/*
+					 * Negative light doesn't make sense.
+					 */
+					if (new_value < 0)
+						new_value = 0;
 
-				approximation->set_pixel(i, j, new_value);
+					approximation->chan(i, j, k) = new_value;
+				}
 			}
 
-                        delete correction;
-			delete correction_count;
+			delete correction;
 
-			if (inc)
+			if (inc) {
+				ui::get()->ip_write();
 				image_rw::output(approximation);
+			}
 
-                        fprintf(stderr, ".");
+			ui::get()->ip_step_done();
 
                 }
 
@@ -786,7 +1249,9 @@ protected:
 
 public:
 
-        ipc(render *input, unsigned int iterations, int _inc, psf *lresponse, psf *nlresponse, int exposure_register) {
+        ipc(render *input, unsigned int iterations, int _inc, psf *lresponse, 
+			psf *nlresponse, int exposure_register,
+			int use_weighted_median, double ipwl) {
                 this->input = input;
                 done = 0;
 		inc = _inc;
@@ -794,6 +1259,8 @@ public:
 		this->lresponse = lresponse;
 		this->nlresponse = nlresponse;
 		this->exposure_register = exposure_register;
+		this->use_weighted_median = use_weighted_median;
+		this->weight_limit = ipwl;
         }
 
         const image *get_image() {
@@ -818,12 +1285,11 @@ public:
 
         virtual int sync() {
 		input->sync();
-                fprintf(stderr, "Iterating Irani-Peleg");
+		ui::get()->ip_start();
                 done = 1;
-                approximation = input->get_image()->clone("IPC Approximation");
+                approximation = optimizations::get_ip_working_image(input->get_image());
                 _ip();
-
-		fprintf(stderr, "\n");
+		ui::get()->ip_done();
 
                 return 0;
         }
@@ -831,6 +1297,8 @@ public:
 	virtual ~ipc() {
 	}
 
+	void free_memory() {
+	}
 };
 
 #endif
