@@ -255,7 +255,7 @@ public:
 	 * Get scale factor.
 	 */
 
-	ale_pos scale() {
+	ale_pos scale() const {
 		return scale_factor;
 	}
 
@@ -428,6 +428,7 @@ public:
 	 * Projective/euclidean inverse followed by barrel distortion.
 	 */
 	struct point scaled_inverse_transform(struct point p) const {
+		assert (p.defined());
 		point q = unscaled_inverse_transform(p);
 
 		q[0] *= scale_factor;
@@ -506,6 +507,19 @@ public:
 
 	}
 
+	/*
+	 * Rotate about a given point in the original reference frame.
+	 */
+	void eu_rotate_about_scaled(point center, ale_pos diff) {
+		assert(center.defined());
+		point fixpoint = scaled_inverse_transform(center);
+		eu_modify(2, diff);
+		point offset = center - transform_scaled(fixpoint);
+
+		eu_modify(0, offset[0]);
+		eu_modify(1, offset[1]);
+	}
+
 	/*
 	 * Modify all euclidean parameters at once.
 	 */
@@ -553,6 +567,32 @@ public:
 
 	}
 
+	/*
+	 * Modify a projective transform according to the group operation.
+	 */
+	void gr_modify(int i1, int i2, ale_pos diff) {
+		assert (_is_projective);
+		assert (i1 == 0 || i1 == 1);
+
+		point diff_vector = (i1 == 0) 
+			          ? point(diff, 0)
+				  : point(0, diff);
+
+		transformation t = *this;
+
+		t.resultant_memo = 0;
+		t.resultant_inverse_memo = 0;
+		t.input_height = (unsigned int) scaled_height();
+		t.input_width  = (unsigned int) scaled_width();
+		t.scale_factor = 1;
+		t.bdcnum = 0;
+
+		resultant_memo = 0;
+		resultant_inverse_memo = 0;
+
+		x[i2] = t.transform_scaled(t.scaled_inverse_transform(x[i2]) + diff_vector);
+	}
+
 	/*
 	 * Modify all projective parameters at once.
 	 */
@@ -568,6 +608,11 @@ public:
 
 	}
 
+	void gpt_set(point x1, point x2, point x3, point x4) {
+		point x[4] = {x1, x2, x3, x4};
+		gpt_set(x);
+	}
+
 	/*
 	 * Get the specified projective parameter
 	 */
@@ -607,6 +652,36 @@ public:
 		}
 	}
 
+	/*
+	 * Rotate by a given amount about a given point.
+	 */
+	void rotate(point p, ale_pos degrees) {
+
+		ale_pos radians = degrees * M_PI / (double) 180;
+
+		if (_is_projective)
+		for (int i = 0; i <= 4; i++) {
+			x[i] -= p;
+			x[i] = point(
+				x[i][0] * cos(radians) + x[i][1] * sin(radians),
+				x[i][1] * cos(radians) - x[i][0] * sin(radians));
+			x[i] += p;
+		} else {
+			point usi = unscaled_inverse_transform(p);
+
+			eu[2] += degrees;
+			resultant_memo = 0;
+			
+			point p_rot = transform_unscaled(usi);
+
+			eu[0] -= (p_rot[0] - p[0]) / scale_factor;
+			eu[1] -= (p_rot[1] - p[1]) / scale_factor;
+		}
+
+		resultant_memo = 0;
+		resultant_inverse_memo = 0;
+	}
+
 	/*
 	 * Set the specified barrel distortion parameter.
 	 */
@@ -1053,6 +1128,25 @@ public:
 
 	}
 
+	void debug_output() {
+		fprintf(stderr, "[t.do ih=%u, iw=%d x=[[%f %f] [%f %f] [%f %f] [%f %f]] eu=[%f %f %f]\n"
+				"      a-f=[%f %f %f %f %f %f %f %f] _a-_f=[%f %f %f %f %f %f %f %f]\n"
+				"      bdcnm=%d ip=%d rm=%d rim=%d sf=%f]\n", 
+				input_height, input_width, 
+				x[0][0], x[0][1], 
+				x[1][0], x[1][1], 
+				x[2][0], x[2][1], 
+				x[3][0], x[3][1], 
+				eu[0], eu[1], eu[2],
+				a, b, c, d, e, f, g, h,
+				_a, _b, _c, _d, _e, _f, _g, _h,
+				bdcnum,
+				_is_projective,
+				resultant_memo,
+				resultant_inverse_memo,
+				scale_factor);
+	}
+
 };
 
 #endif
